@@ -1,26 +1,29 @@
-import {
-  from,
-  lastValueFrom,
-  merge,
-  MonoTypeOperatorFunction,
-  Observable,
-  of,
-  Subject,
-  Subscription,
-  throwError,
-} from 'rxjs';
+import FingerprintJS from '@fingerprintjs/fingerprintjs';
+import { from, lastValueFrom, MonoTypeOperatorFunction, Observable, Subject, Subscription, throwError } from 'rxjs';
 import { fromFetch } from 'rxjs/fetch';
-import { catchError, filter, map, mergeMap, retryWhen, share, takeUntil, tap, throwIfEmpty } from 'rxjs/operators';
+import {
+  catchError,
+  filter,
+  finalize,
+  map,
+  mergeMap,
+  retryWhen,
+  share,
+  takeUntil,
+  tap,
+  throwIfEmpty,
+} from 'rxjs/operators';
 import { v4 as uuidv4 } from 'uuid';
 
-import { AppEvents, DataQueryErrorType } from '@grafana/data';
+import { AppEvents, DataQueryErrorType, deprecationWarning } from '@grafana/data';
 import { BackendSrv as BackendService, BackendSrvRequest, config, FetchError, FetchResponse } from '@grafana/runtime';
 import appEvents from 'app/core/app_events';
 import { getConfig } from 'app/core/config';
+import { getSessionExpiry, hasSessionExpiry } from 'app/core/utils/auth';
 import { loadUrlToken } from 'app/core/utils/urlToken';
+import { getDashboardAPI } from 'app/features/dashboard/api/dashboard_api';
 import { DashboardModel } from 'app/features/dashboard/state';
 import { DashboardSearchItem } from 'app/features/search/types';
-import { getGrafanaStorage } from 'app/features/storage/storage';
 import { TokenRevokedModal } from 'app/features/users/TokenRevokedModal';
 import { DashboardDTO, FolderDTO } from 'app/types';
 
@@ -51,13 +54,22 @@ export interface FolderRequestOptions {
   withAccessControl?: boolean;
 }
 
+const GRAFANA_TRACEID_HEADER = 'grafana-trace-id';
+
+export interface InspectorStream {
+  response: FetchResponse | FetchError;
+  requestId?: string;
+}
+
 export class BackendSrv implements BackendService {
   private inFlightRequests: Subject<string> = new Subject<string>();
   private HTTP_REQUEST_CANCELED = -1;
   private noBackendCache: boolean;
-  private inspectorStream: Subject<FetchResponse | FetchError> = new Subject<FetchResponse | FetchError>();
+  private inspectorStream: Subject<InspectorStream> = new Subject<InspectorStream>();
   private readonly fetchQueue: FetchQueue;
   private readonly responseQueue: ResponseQueue;
+  private _tokenRotationInProgress?: Observable<FetchResponse> | null = null;
+  private deviceID?: string | null = null;
 
   private dependencies: BackendSrvDependencies = {
     fromFetch: fromFetch,
@@ -65,7 +77,6 @@ export class BackendSrv implements BackendService {
     contextSrv: contextSrv,
     logout: () => {
       contextSrv.setLoggedOut();
-      window.location.reload();
     },
   };
 
@@ -81,10 +92,23 @@ export class BackendSrv implements BackendService {
     this.internalFetch = this.internalFetch.bind(this);
     this.fetchQueue = new FetchQueue();
     this.responseQueue = new ResponseQueue(this.fetchQueue, this.internalFetch);
+
+    this.initGrafanaDeviceID();
+
     new FetchQueueWorker(this.fetchQueue, this.responseQueue, getConfig());
   }
 
-  async request<T = any>(options: BackendSrvRequest): Promise<T> {
+  private async initGrafanaDeviceID() {
+    try {
+      const fp = await FingerprintJS.load();
+      const result = await fp.get();
+      this.deviceID = result.visitorId;
+    } catch (error) {
+      console.error(error);
+    }
+  }
+
+  async request<T = unknown>(options: BackendSrvRequest): Promise<T> {
     return await lastValueFrom(this.fetch<T>(options).pipe(map((response: FetchResponse<T>) => response.data)));
   }
 
@@ -132,27 +156,20 @@ export class BackendSrv implements BackendService {
 
     const token = loadUrlToken();
     if (token !== null && token !== '') {
-      if (!options.headers) {
-        options.headers = {};
-      }
-
       if (config.jwtUrlLogin && config.jwtHeaderName) {
+        options.headers = options.headers ?? {};
         options.headers[config.jwtHeaderName] = `${token}`;
       }
     }
 
-    const fromFetchStream = this.getFromFetchStream<T>(options);
-    const failureStream = fromFetchStream.pipe(this.toFailureStream<T>(options));
-    const successStream = fromFetchStream.pipe(
-      filter((response) => response.ok === true),
-      tap((response) => {
-        this.showSuccessAlert(response);
-        this.inspectorStream.next(response);
-      })
-    );
+    if (!!this.deviceID) {
+      options.headers = options.headers ?? {};
+      options.headers['X-Grafana-Device-Id'] = `${this.deviceID}`;
+    }
 
-    return merge(successStream, failureStream).pipe(
-      catchError((err: FetchError) => throwError(this.processRequestError(options, err))),
+    return this.getFromFetchStream<T>(options).pipe(
+      this.handleStreamResponse<T>(options),
+      this.handleStreamError(options),
       this.handleStreamCancellation(options)
     );
   }
@@ -165,8 +182,8 @@ export class BackendSrv implements BackendService {
     this.inFlightRequests.next(CANCEL_ALL_REQUESTS_REQUEST_ID);
   }
 
-  async datasourceRequest(options: BackendSrvRequest): Promise<any> {
-    return lastValueFrom(this.fetch(options));
+  async datasourceRequest<T = unknown>(options: BackendSrvRequest) {
+    return lastValueFrom(this.fetch<T>(options));
   }
 
   private parseRequestOptions(options: BackendSrvRequest): BackendSrvRequest {
@@ -225,61 +242,13 @@ export class BackendSrv implements BackendService {
           type,
           redirected,
           config: options,
+          traceId: response.headers.get(GRAFANA_TRACEID_HEADER) ?? undefined,
         };
         return fetchResponse;
-      }),
-      share() // sharing this so we can split into success and failure and then merge back
+      })
     );
   }
 
-  private toFailureStream<T>(options: BackendSrvRequest): MonoTypeOperatorFunction<FetchResponse<T>> {
-    const { isSignedIn } = this.dependencies.contextSrv.user;
-
-    return (inputStream) =>
-      inputStream.pipe(
-        filter((response) => response.ok === false),
-        mergeMap((response) => {
-          const { status, statusText, data } = response;
-          const fetchErrorResponse: FetchError = { status, statusText, data, config: options };
-          return throwError(fetchErrorResponse);
-        }),
-        retryWhen((attempts: Observable<any>) =>
-          attempts.pipe(
-            mergeMap((error, i) => {
-              const firstAttempt = i === 0 && options.retry === 0;
-
-              if (error.status === 401 && isLocalUrl(options.url) && firstAttempt && isSignedIn) {
-                if (error.data?.error?.id === 'ERR_TOKEN_REVOKED') {
-                  this.dependencies.appEvents.publish(
-                    new ShowModalReactEvent({
-                      component: TokenRevokedModal,
-                      props: {
-                        maxConcurrentSessions: error.data?.error?.maxConcurrentSessions,
-                      },
-                    })
-                  );
-
-                  return of({});
-                }
-
-                return from(this.loginPing()).pipe(
-                  catchError((err) => {
-                    if (err.status === 401) {
-                      this.dependencies.logout();
-                      return throwError(err);
-                    }
-                    return throwError(err);
-                  })
-                );
-              }
-
-              return throwError(error);
-            })
-          )
-        )
-      );
-  }
-
   showApplicationErrorAlert(err: FetchError) {}
 
   showSuccessAlert<T>(response: FetchResponse<T>) {
@@ -289,7 +258,7 @@ export class BackendSrv implements BackendService {
       return;
     }
 
-    // is showSuccessAlert is undefined we only show alerts non GET request, non data query and local api requests
+    // if showSuccessAlert is undefined we only show alerts non GET request, non data query and local api requests
     if (
       config.showSuccessAlert === undefined &&
       (config.method === 'GET' || isDataQuery(config.url) || !isLocalUrl(config.url))
@@ -317,6 +286,11 @@ export class BackendSrv implements BackendService {
     let description = '';
     let message = err.data.message;
 
+    // Sometimes we have a better error message on err.message
+    if (message === 'Unexpected error' && err.message) {
+      message = err.message;
+    }
+
     if (message.length > 80) {
       description = message;
       message = 'Error';
@@ -365,11 +339,85 @@ export class BackendSrv implements BackendService {
       }, 50);
     }
 
-    this.inspectorStream.next(err);
+    this.inspectorStream.next({ response: err, requestId: options.requestId });
     return err;
   }
 
-  private handleStreamCancellation(options: BackendSrvRequest): MonoTypeOperatorFunction<FetchResponse<any>> {
+  private handleStreamResponse<T>(options: BackendSrvRequest): MonoTypeOperatorFunction<FetchResponse<T>> {
+    return (inputStream) =>
+      inputStream.pipe(
+        map((response) => {
+          if (!response.ok) {
+            const { status, statusText, data } = response;
+            const fetchErrorResponse: FetchError = {
+              status,
+              statusText,
+              data,
+              config: options,
+              traceId: response.headers.get(GRAFANA_TRACEID_HEADER) ?? undefined,
+            };
+            throw fetchErrorResponse;
+          }
+          return response;
+        }),
+        tap((response) => {
+          this.showSuccessAlert(response);
+          this.inspectorStream.next({ response: response, requestId: options.requestId });
+        })
+      );
+  }
+
+  private handleStreamError<T>(options: BackendSrvRequest): MonoTypeOperatorFunction<FetchResponse<T>> {
+    const { isSignedIn } = this.dependencies.contextSrv.user;
+
+    return (inputStream) =>
+      inputStream.pipe(
+        retryWhen((attempts) =>
+          attempts.pipe(
+            mergeMap((error, i) => {
+              const firstAttempt = i === 0 && options.retry === 0;
+
+              if (error.status === 401 && isLocalUrl(options.url) && firstAttempt && isSignedIn) {
+                if (error.data?.error?.id === 'ERR_TOKEN_REVOKED') {
+                  this.dependencies.appEvents.publish(
+                    new ShowModalReactEvent({
+                      component: TokenRevokedModal,
+                      props: {
+                        maxConcurrentSessions: error.data?.error?.maxConcurrentSessions,
+                      },
+                    })
+                  );
+                  return throwError(() => error);
+                }
+
+                let authChecker = this.loginPing();
+                if (hasSessionExpiry()) {
+                  const expired = getSessionExpiry() * 1000 < Date.now();
+                  if (expired) {
+                    authChecker = this.rotateToken();
+                  }
+                }
+
+                return from(authChecker).pipe(
+                  catchError((err) => {
+                    if (err.status === 401) {
+                      this.dependencies.logout();
+                      return throwError(err);
+                    }
+                    return throwError(err);
+                  })
+                );
+              }
+
+              return throwError(error);
+            })
+          )
+        ),
+        catchError((err: FetchError) => throwError(() => this.processRequestError(options, err)))
+      );
+  }
+
+  private handleStreamCancellation(options: BackendSrvRequest): MonoTypeOperatorFunction<FetchResponse> {
     return (inputStream) =>
       inputStream.pipe(
         takeUntil(
@@ -405,7 +453,7 @@ export class BackendSrv implements BackendService {
       );
   }
 
-  getInspectorStream(): Observable<FetchResponse<any> | FetchError> {
+  getInspectorStream(): Observable<InspectorStream> {
     return this.inspectorStream;
   }
 
@@ -418,55 +466,68 @@ export class BackendSrv implements BackendService {
     return this.request<T>({ ...options, method: 'GET', url, params, requestId });
   }
 
-  async delete<T = any>(url: string, data?: any, options?: Partial<BackendSrvRequest>) {
+  async delete<T = unknown>(url: string, data?: unknown, options?: Partial<BackendSrvRequest>) {
     return this.request<T>({ ...options, method: 'DELETE', url, data });
   }
 
-  async post<T = any>(url: string, data?: any, options?: Partial<BackendSrvRequest>) {
+  async post<T = any>(url: string, data?: unknown, options?: Partial<BackendSrvRequest>) {
     return this.request<T>({ ...options, method: 'POST', url, data });
   }
 
-  async patch<T = any>(url: string, data: any, options?: Partial<BackendSrvRequest>) {
+  async patch<T = any>(url: string, data: unknown, options?: Partial<BackendSrvRequest>) {
     return this.request<T>({ ...options, method: 'PATCH', url, data });
   }
 
-  async put<T = any>(url: string, data: any, options?: Partial<BackendSrvRequest>): Promise<T> {
+  async put<T = any>(url: string, data: unknown, options?: Partial<BackendSrvRequest>): Promise<T> {
     return this.request<T>({ ...options, method: 'PUT', url, data });
   }
 
-  withNoBackendCache(callback: any) {
+  withNoBackendCache(callback: () => Promise<void>) {
     this.noBackendCache = true;
     return callback().finally(() => {
       this.noBackendCache = false;
     });
   }
 
+  rotateToken() {
+    if (this._tokenRotationInProgress) {
+      return this._tokenRotationInProgress;
+    }
+
+    this._tokenRotationInProgress = this.fetch({ url: '/api/user/auth-tokens/rotate', method: 'POST', retry: 1 }).pipe(
+      finalize(() => {
+        this._tokenRotationInProgress = null;
+      }),
+      share()
+    );
+
+    return this._tokenRotationInProgress;
+  }
+
   loginPing() {
-    return this.request({ url: '/api/login/ping', method: 'GET', retry: 1 });
+    return this.fetch({ url: '/api/login/ping', method: 'GET', retry: 1 });
   }
 
   /** @deprecated */
-  search(query: any): Promise<DashboardSearchItem[]> {
+  search(query: Parameters<typeof this.get>[1]): Promise<DashboardSearchItem[]> {
     return this.get('/api/search', query);
   }
 
+  /** @deprecated */
   getDashboardByUid(uid: string): Promise<DashboardDTO> {
-    if (uid.indexOf('/') > 0 && config.featureToggles.dashboardsFromStorage) {
-      return getGrafanaStorage().getDashboard(uid);
-    }
-    return this.get<DashboardDTO>(`/api/dashboards/uid/${uid}`);
+    // NOTE: When this is removed, we can also remove most instances of:
+    // jest.mock('app/features/live/dashboard/dashboardWatcher
+    deprecationWarning('backend_srv', 'getDashboardByUid(uid)', 'getDashboardAPI().getDashboardDTO(uid)');
+    return getDashboardAPI().getDashboardDTO(uid);
   }
 
-  validateDashboard(dashboard: DashboardModel) {
-    // We want to send the dashboard as a JSON string (in the JSON body payload) so we can get accurate error line numbers back
-    const dashboardJson = JSON.stringify(dashboard, replaceJsonNulls, 2);
-
-    return this.request<ValidateDashboardResponse>({
-      method: 'POST',
-      url: `/api/dashboards/validate`,
-      data: { dashboard: dashboardJson },
-      showSuccessAlert: false,
-      showErrorAlert: false,
+  validateDashboard(dashboard: DashboardModel): Promise<ValidateDashboardResponse> {
+    // support for this function will be implemented in the k8s flavored api-server
+    // hidden by experimental feature flag:
+    //  config.featureToggles.showDashboardValidationWarnings
+    return Promise.resolve({
+      isValid: false,
+      message: 'dashboard validation is not supported',
     });
   }
 
@@ -480,7 +541,9 @@ export class BackendSrv implements BackendService {
       queryParams.set('accesscontrol', 'true');
     }
 
-    return this.get<FolderDTO>(`/api/folders/${uid}?${queryParams.toString()}`);
+    return this.get<FolderDTO>(`/api/folders/${uid}?${queryParams.toString()}`, undefined, undefined, {
+      showErrorAlert: false,
+    });
   }
 }
 
@@ -492,10 +555,3 @@ interface ValidateDashboardResponse {
   isValid: boolean;
   message?: string;
 }
-
-function replaceJsonNulls<T extends unknown>(key: string, value: T): T | undefined {
-  if (typeof value === 'number' && !Number.isFinite(value)) {
-    return undefined;
-  }
-  return value;
-}
