@@ -11,7 +11,7 @@ import './jquery.flot.events';
 import $ from 'jquery';
 import { clone, find, flatten, isUndefined, map, max as _max, min as _min, sortBy as _sortBy, toNumber } from 'lodash';
 import React from 'react';
-import ReactDOM from 'react-dom';
+import { createRoot, Root } from 'react-dom/client';
 
 import {
   DataFrame,
@@ -36,7 +36,7 @@ import {
   PanelEvents,
   toUtc,
 } from '@grafana/data';
-import { graphTickFormatter, graphTimeFormat, MenuItemProps, MenuItemsGroup } from '@grafana/ui';
+import { MenuItemProps, MenuItemsGroup } from '@grafana/ui';
 import { coreModule } from 'app/angular/core_module';
 import config from 'app/core/config';
 import { updateLegendValues } from 'app/core/core';
@@ -44,10 +44,9 @@ import { ContextSrv } from 'app/core/services/context_srv';
 import { provideTheme } from 'app/core/utils/ConfigProvider';
 import { tickStep } from 'app/core/utils/ticks';
 import { TimeSrv } from 'app/features/dashboard/services/TimeSrv';
+import { DashboardModel } from 'app/features/dashboard/state';
 import { getFieldLinksSupplier } from 'app/features/panel/panellinks/linkSuppliers';
 
-import { DashboardModel } from '../../../features/dashboard/state';
-
 import { GraphContextMenuCtrl } from './GraphContextMenuCtrl';
 import { GraphLegendProps, Legend } from './Legend/Legend';
 import { alignYLevel } from './align_yaxes';
@@ -57,7 +56,7 @@ import { convertToHistogramData } from './histogram';
 import { GraphCtrl } from './module';
 import { ThresholdManager } from './threshold_manager';
 import { TimeRegionManager } from './time_region_manager';
-import { isLegacyGraphHoverEvent } from './utils';
+import { isLegacyGraphHoverEvent, graphTickFormatter, graphTimeFormat } from './utils';
 
 const LegendWithThemeProvider = provideTheme(Legend, config.theme2);
 
@@ -76,6 +75,7 @@ class GraphElement {
   thresholdManager: ThresholdManager;
   timeRegionManager: TimeRegionManager;
   declare legendElem: HTMLElement;
+  declare legendElemRoot: Root;
 
   constructor(
     private scope: any,
@@ -118,6 +118,7 @@ class GraphElement {
     // get graph legend element
     if (this.elem && this.elem.parent) {
       this.legendElem = this.elem.parent().find('.graph-legend')[0];
+      this.legendElemRoot = createRoot(this.legendElem);
     }
   }
 
@@ -134,9 +135,14 @@ class GraphElement {
 
     if (!this.panel.legend.show) {
       if (this.legendElem.hasChildNodes()) {
-        ReactDOM.unmountComponentAtNode(this.legendElem);
+        this.legendElemRoot.render(null);
       }
-      this.renderPanel();
+      // we need to wait for react to finish rendering the legend before we can render the graph
+      // this is a slightly worse version of the `renderCallback` logic we use below
+      // the problem here is there's nothing to pass a `renderCallback` to since we don't want to render the legend at all.
+      setTimeout(() => {
+        this.renderPanel();
+      });
       return;
     }
 
@@ -153,10 +159,13 @@ class GraphElement {
       onToggleSort: this.ctrl.onToggleSort,
       onColorChange: this.ctrl.onColorChange,
       onToggleAxis: this.ctrl.onToggleAxis,
+      renderCallback: this.renderPanel.bind(this),
     };
 
     const legendReactElem = React.createElement(LegendWithThemeProvider, legendProps);
-    ReactDOM.render(legendReactElem, this.legendElem, () => this.renderPanel());
+
+    // render callback isn't supported in react 18+, see: https://github.com/reactwg/react-18/discussions/5
+    this.legendElemRoot.render(legendReactElem);
   }
 
   onGraphHover(evt: LegacyGraphHoverEventPayload | DataHoverPayload) {
@@ -192,7 +201,7 @@ class GraphElement {
     this.elem.off();
     this.elem.remove();
 
-    ReactDOM.unmountComponentAtNode(this.legendElem);
+    this.legendElemRoot.unmount();
   }
 
   onGraphHoverClear(handler: LegacyEventHandler<any>) {
@@ -316,7 +325,7 @@ class GraphElement {
           field: { config: fieldConfig, type: FieldType.number },
           theme: config.theme2,
           timeZone: this.dashboard.getTimezone(),
-        })(field.values.get(dataIndex));
+        })(field.values[dataIndex]);
         linksSupplier = links.length
           ? getFieldLinksSupplier({
               display: fieldDisplay,
@@ -360,13 +369,13 @@ class GraphElement {
       return dataIndex;
     }
 
-    const field = timeField.values.get(dataIndex);
+    const field = timeField.values[dataIndex];
 
     if (field === ts) {
       return dataIndex;
     }
 
-    const correctIndex = timeField.values.toArray().findIndex((value) => value === ts);
+    const correctIndex = timeField.values.findIndex((value) => value === ts);
     return correctIndex > -1 ? correctIndex : dataIndex;
   }
 
